from typing import List, Tuple, Union

import librosa
import numpy as np
import soundfile as sf

ndarray = np.ndarray
genlist = Union[ndarray, list]


# Helper functions:
def secondsToSamples(seconds: float, sr: int) -> int:
    return int(seconds*sr)


def maxIndex(l: genlist) -> int:
    maxI = 0
    maxVal = l[0]
    for i in range(1, len(l)):
        val = l[i]
        if val > maxVal:
            maxVal = val
            maxI = i
    return maxI


def chunkAverage(l: genlist, chunkSize: int) -> genlist:
    if chunkSize > len(l):
        raise ValueError("ChunkSize must be less than length of list.")
    avgList = []
    currentIndex = 0
    while (currentIndex + chunkSize) < len(l):
        avgList.append(np.mean(l[currentIndex:currentIndex + chunkSize]))
        currentIndex += chunkSize
    avgList.append(np.mean(l[currentIndex:]))
    return avgList


def loadFromFile(fileName: str) -> Tuple[np.ndarray, int]:
    wf, sr = sf.read(fileName, always_2d=True)
    return (wf, sr)


def getNSecondsFromWaveForm(
        waveForm: np.ndarray,
        sr: int,
        seconds: float
        ) -> np.ndarray:
    return waveForm[:secondsToSamples(seconds, sr)]


# Main functions:
class Analysis():
    @staticmethod
    def condenseToSingleChannel(waveForm: np.ndarray) -> np.ndarray:
        """Condenses multi-channel waveforms down to one channel.

        This is done by taking the first channel value, no other black magic.
        Parameters
        ----------
        waveForm : np.ndarray
            A waveform that might be multi channel.

        Returns
        -------
        np.ndarray
            The waveform condensed to a single channel.

        """
        shape = waveForm.shape
        if len(shape) == 2 and waveForm.shape[1] == 2:
            retArr = np.array(
                list(map(lambda l: l[0], waveForm))
            )
            return retArr
        return waveForm

    @staticmethod
    def rfftAudible(waveForm: np.ndarray) -> np.ndarray:
        """Take a (real-valued) Fast Fourier transform of a waveform.

        Parameters
        ----------
        waveForm : np.ndarray
            Single Channel waveform.

        Returns
        -------
        np.ndarray
            waveForm transformed into the (audible) frequency domain.
        """
        fft = np.fft.rfft(waveForm)
        mfft = []
        for x in fft:
            if x <= 200 or x >= 2000:
                mfft.append(0)
            else:
                mfft.append(abs(x))
        return np.array(mfft)

    @staticmethod
    def startingNote(waveForm: ndarray, sr: int) -> int:
        """Get's the starting note of a waveform.

        Parameters
        ----------
        waveForm : ndarray
            Any ol' waveform.
        sr : int
            Sample rate of the waveform.

        Returns
        -------
        int
            The most dominant frequency of waveForm in the first 1.5 seconds.

        """
        sc = Analysis.condenseToSingleChannel(waveForm)
        beatSample = getNSecondsFromWaveForm(sc, sr, 1.5)
        trimmed, _ = librosa.effects.trim(beatSample)
        adRFFT = Analysis.rfftAudible(trimmed)
        return maxIndex(adRFFT)

    @staticmethod
    def startingNoteFromFile(fileName: str) -> int:
        """Convienient wrapper for startingNote to load file automatically.

        For more info see startingNote
        Parameters
        ----------
        fileName : str
            fileName of wavFile.

        Returns
        -------
        int
            Dominant frequency present in fileName.

        """
        waveForm, sr = loadFromFile(fileName)
        return Analysis.startingNote(waveForm, sr)



import math
from typing import List, Tuple, TypeVar

# Global Types
noteTupleDuration = Tuple[str, int, float, int]
noteTuple = Tuple[str, int, float]
frequencyTuple = Tuple[float, float]
T = TypeVar('T')
# Note tuple has (Note Letter, Octave, Duration(, repetition times))

# Global Variables
currentOctave = 4
shiftDict = {
    "A": 0,
    "A#": 1,
    "Bf": 1,
    "B": 2,
    "C": -9,
    "C#": -8,
    "Df": -8,
    "D": -7,
    "D#": -6,
    "Ef": -6,
    "E": -5,
    "F": -4,
    "F#": -3,
    "Gf": -3,
    "G": -2,
    "G#": -1,
    "Af": -1
}

baseFreq = 440  # A4 is 440
expBase = math.pow(2, 1.0/12)


# Helper functions:
def isFloat(s: str) -> bool:
    return s.replace(".", "").isdigit()


def appendMultipleTimes(l: List[T], a: T, i: int) -> list:
    c = 0
    nl = l
    while c < i:
        nl.append(a)
        c += 1
    return nl


def lastCharInString(s: str) -> str:
    return s[len(s)-1]


def getArrayFromFile(fileName: str) -> list:
    with open(fileName, "r") as file:
        return list(map(lambda s: s.replace("\n", ""), file.readlines()))


# Main functions:
class ParseMusic():
    @staticmethod
    def parseNoteString(note: str) -> noteTupleDuration:
        """Takes a string like "2x 3 E5" and parses it into a note tuple.

        Parameters
        ----------
        note: str
            String representation of note.
                e.g. 2x 3 E5 for 2 repetitions of an E in the 5th octave with
                duration 3

        Returns
        -------
        noteTuple
            NoteTuple with information from noteString
                "2x 3 E5" -> (E, 5, 3, 2)

        """

        # TODO: avoid globals
        global currentOctave
        parts = note.split(" ")
        nums = 1
        duration = 1.0
        currIndex = 0
        if(
            len(parts[currIndex]) > 1 and
            isFloat(parts[currIndex][:-1]) and
            parts[currIndex][-1:] == 'x'
        ):
            nums = int(parts[currIndex][:-1])
            currIndex += 1
        if (
            isFloat(parts[currIndex]) and
            lastCharInString(parts[currIndex]) != 'x'
        ):
            duration = float(parts[currIndex])
            currIndex += 1
        noteString = parts[currIndex]
        if isFloat(noteString[len(noteString) - 1]):
            currentOctave = int(noteString[len(noteString) - 1])
            noteString = noteString[:-1]
        return (noteString, currentOctave, duration, nums)

    @staticmethod
    def parseArrayToNotePair(noteArray: List[str]) -> List[List[noteTuple]]:
        """Takes a list of note strings and returns a list of note tuples.

        Parameters
        ----------
        noteArray : List[str]
            List of note strings.
                e.g.
                    [
                        "2x 3 E5",
                        "2 D",
                        "3x C",
                        ...
                    ]

        Returns
        -------
        List[noteTuple]
            Parsed list.
                e.g.
                    [
                        (E, 5, 3, 2),
                        (D, 5, 2, 1),
                        (C, 5, 1, 3),
                        ...
                    ]

        """
        parsedArray = []  # type: List[List[noteTuple]]

        for note in noteArray:
            arr = []
            if "," not in note:
                ns, o, d, n = ParseMusic.parseNoteString(note)
                appendMultipleTimes(parsedArray, [(ns, o, d)], n)
                continue
            else:
                splitArr = note.split(",")
                genDuration = ParseMusic.parseNoteString(splitArr[0])[2]
                for partialNote in splitArr:
                    partialNote = partialNote.strip()
                    ns, o, _, n = ParseMusic.parseNoteString(partialNote)
                    arr.append((ns, o, genDuration))
            parsedArray.append(arr)

        return parsedArray

    @staticmethod
    def noteTupleToFrequency(noteT: noteTuple) -> frequencyTuple:
        """Takes a note tuple and returns a frequency tuple for the note.

        Parameters
        ----------
        notePair : noteTuple
            NoteTuple for note.

        Returns
        -------
        frequencyTuple
            Frequency tuple for note.

        """
        note, octave, duration = noteT
        if note == "_":
            return (0, duration)
        semitoneDiff = (octave-4)*12 + shiftDict[note]
        roundedFreq = round(baseFreq * math.pow(expBase, semitoneDiff), 2)
        return (roundedFreq, duration)

    @staticmethod
    def fileToFrequency(fileName: str) -> List[List[frequencyTuple]]:
        """Short summary.

        Parameters
        ----------
        fileName : str
            Filename of music file.

        Returns
        -------
        List[List[frequencyTuple]]
            Music file parsed into frequencies.

        """
        arr = getArrayFromFile(fileName)
        np = ParseMusic.parseArrayToNotePair(arr)
        song = []
        for chord in np:
            freqChord = list(map(ParseMusic.noteTupleToFrequency, chord))
            song.append(freqChord)
        return song



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="tuning_fork",
    version="1.4",
    author="Garrett Credi",
    author_email="gcc@ameritech.net",
    description="A clip/sample auto tuner",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/ddxtanx/TuningFork",
    packages=["tuning_fork", "tuning_fork.tools"],
    py_modules=["tuning_fork.main"],
    classifiers=(
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ),
    install_requires=["librosa", "numpy", "pysoundfile"]
)



from typing import Tuple

import librosa
import numpy as np
import soundfile as sf
from tuning_fork.tools.analysis import Analysis
from tuning_fork.tools.parseMusic import ParseMusic


def getSPB(bpm: float) -> float:
    bps = bpm/60
    return 1.0/bps


def addList(l1: list, l2: list) -> list:
    if len(l1) != len(l2):
        raise ValueError("Lists must be of Equal Length")
    comb = zip(l1, l2)
    return list(map(lambda t: t[0] + t[1], comb))


def mdArrayAverage(md: list) -> list:
    sumArr = md[0]
    for i in range(1, len(md)):
        sumArr = addList(sumArr, md[i])
    return list(map(lambda v: float(v)/len(md), sumArr))


def shiftPitchBy(
            noteArray: np.ndarray,
            sr: int,
            shift: float
            ) -> np.ndarray:
    return librosa.effects.pitch_shift(noteArray, sr, shift)


def speedUpBy(
            noteArray: np.ndarray,
            sr: int,
            factor: float
            ) -> np.ndarray:
    spedUp = librosa.effects.time_stretch(noteArray, factor)
    return spedUp


class TuningFork():
    @staticmethod
    def speedUpTo(
                waveForm: np.ndarray,
                sr: int,
                shiftTo: float
                ) -> np.ndarray:
        """Speeds up waveform to a certain length in seconds.

        Parameters
        ----------
        waveForm : np.ndarray
            waveForm to be sped up.
        sr : int
            Sample rate of waveForm.
        shiftTo : float
            Number of seconds waveForm is shifted to.

        Returns
        -------
        np.ndarray
            waveForm shifted to shiftTo seconds.

        """
        dur = librosa.core.get_duration(waveForm, sr)
        ratio = dur/shiftTo
        spedUp = speedUpBy(waveForm, sr, ratio)
        return spedUp

    @staticmethod
    def pitchTo(
            waveForm: np.ndarray,
            sr: int,
            shiftTo: int,
            bpm: float
            ) -> np.ndarray:
        """Changes the dominant pitch of a waveForm to a specified frequency.

        Parameters
        ----------
        waveForm : np.ndarray
            Same as above.
        sr : int
            Sample rate of waveform.
        shiftTo : int
            Pitch frequency to be shifted to.
        bpm : float
            If you are shifting to fit to a song, this is the bpm of the song.

        Returns
        -------
        np.ndarray
            Description of returned object.

        """
        spb = getSPB(bpm)

        currentFrequency = Analysis.startingNote(waveForm, sr)
        freqRatio = shiftTo/float(currentFrequency)
        shift = np.log2(freqRatio)
        shiftedPitchWF = shiftPitchBy(waveForm, sr, 12*shift)
        return TuningFork.speedUpTo(shiftedPitchWF, sr, spb)

    @staticmethod
    def emptyBeatTrack(
            likeWaveForm: np.ndarray,
            sr: int,
            bpm: float
            ) -> np.ndarray:
        """Creates an (almost) empty track section.

        Used for making rests.
        Parameters
        ----------
        likeWaveForm : np.ndarray
            The waveForm the track should be modelled after.
        sr : int
            The sr of likeWaveForm.
        bpm : float
            Same as bpm from pitchTo.

        Returns
        -------
        np.ndarray
            A silent track similar to likeWaveForm.

        """
        spb = getSPB(bpm)
        emptyLike = np.full_like(likeWaveForm, .01)
        return TuningFork.speedUpTo(emptyLike, sr, spb)

    @staticmethod
    def sampleIntoSong(
            sampleTrack: np.ndarray,
            sr: int,
            frequencyList: list,
            bpm: float
            ) -> np.ndarray:
        """Takes a waveForm and autotunes it to a list of frequencies.

        Parameters
        ----------
        sampleTrack : np.ndarray
            The track to be sampled into the song.
        sr : int
            The sample rate of the sample track.
        frequencyList : list
            The list of frequencies to fit sampleTrack to.
        bpm : float
            Same as other bpm.

        Returns
        -------
        np.ndarray
            A waveform that is sampleTrack fitted to frequencyList.

        """
        song = np.array([])
        for freqList in frequencyList:
            nextParts = []
            for freq, duration in freqList:
                if freq != 0:
                    pitched = TuningFork.pitchTo(sampleTrack, sr, freq, bpm)
                    nextParts.append(pitched)
                else:
                    empty = TuningFork.emptyBeatTrack(sampleTrack, sr, bpm)
                    nextParts.append(empty)
            avgArr = np.array(mdArrayAverage(nextParts))
            spedUpPart = speedUpBy(avgArr, sr, 1.0/duration)
            song = np.concatenate((song, spedUpPart))
        return song

    @staticmethod
    def sampleWAVFileIntoMusic(
                        wavFileName: str,
                        musicFileName: str,
                        bpm: float
                        ) -> Tuple[np.ndarray, int]:
        """Wrapper of sampleIntoSong for convienience.

        Parameters
        ----------
        wavFileName : str
            The filename of the wav sample file.
        musicFileName : str
            The filename of the music template file.
        bpm : float
            same as above.

        Returns
        -------
        Tuple[np.ndarray, int]
            A tuple consisting of the fitted song and the sample rate of it.

        """
        waveForm, sr = sf.read(wavFileName)

        waveForm, _ = librosa.effects.trim(waveForm)
        waveForm = Analysis.condenseToSingleChannel(waveForm)
        freqList = ParseMusic.fileToFrequency(musicFileName)

        wP = TuningFork.sampleIntoSong(waveForm, sr, freqList, bpm=bpm)
        return (wP, sr)
